@@ -51,11 +51,14 @@
 #define REST_URL "https://secure.smugmug.com/services/api/rest/1.2.2/"
 #define UPLOAD_URL "http://upload.smugmug.com/photos/xmladd.mg"
 
+static void initable_iface_init (gpointer g_iface, gpointer iface_data);
 static void collections_iface_init (gpointer g_iface, gpointer iface_data);
 static void photo_upload_iface_init (gpointer g_iface, gpointer iface_data);
 static void video_upload_iface_init (gpointer g_iface, gpointer iface_data);
 
 G_DEFINE_TYPE_WITH_CODE (SwServiceSmugmug, sw_service_smugmug, SW_TYPE_SERVICE,
+                         G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE,
+                                                initable_iface_init)
                          G_IMPLEMENT_INTERFACE (SW_TYPE_COLLECTIONS_IFACE,
                                                 collections_iface_init)
                          G_IMPLEMENT_INTERFACE (SW_TYPE_PHOTO_UPLOAD_IFACE,
@@ -63,11 +66,6 @@ G_DEFINE_TYPE_WITH_CODE (SwServiceSmugmug, sw_service_smugmug, SW_TYPE_SERVICE,
                          G_IMPLEMENT_INTERFACE (SW_TYPE_VIDEO_UPLOAD_IFACE,
                                                 video_upload_iface_init));
 
-#define GET_PRIVATE(o) \
-  (G_TYPE_INSTANCE_GET_PRIVATE ((o), SW_TYPE_SERVICE_SMUGMUG, SwServiceSmugmugPrivate))
-
-#define ALBUM_PREFIX "smugmug-"
-
 struct _SwServiceSmugmugPrivate {
   const gchar *api_key;
   const gchar *api_secret;
@@ -75,6 +73,22 @@ struct _SwServiceSmugmugPrivate {
   RestProxy *rest_proxy;
   RestProxy *upload_proxy;
   RestProxy *auth_proxy;
+
+  gboolean configured;
+  gboolean authorized;
+  gboolean inited;
+};
+
+#define ALBUM_PREFIX "smugmug-"
+
+static const ParameterNameMapping upload_params[] = {
+  { "title", "Caption" },
+  { "x-smugmug-hidden", "Hidden" },
+  { "x-smugmug-altitude", "Altitude" },
+  { "x-smugmug-latitude", "Latitude" },
+  { "x-smugmug-longitude", "Longitude" },
+  { "x-smugmug-keywords", "Keywords" },
+  { NULL, NULL }
 };
 
 enum {
@@ -96,6 +110,7 @@ get_static_caps (SwService *service)
     HAS_PHOTO_UPLOAD_IFACE,
     HAS_VIDEO_UPLOAD_IFACE,
     HAS_BANISHABLE_IFACE,
+    HAS_COLLECTIONS_IFACE,
 
     NULL
   };
@@ -103,13 +118,116 @@ get_static_caps (SwService *service)
   return caps;
 }
 
-static void
-got_tokens_cb (RestProxy *proxy, gboolean authorised, gpointer user_data)
+static const char **
+get_dynamic_caps (SwService *service)
 {
-  SwServiceSmugmug *self = (SwServiceSmugmug *) user_data;
+  SwServiceSmugmugPrivate *priv = SW_SERVICE_SMUGMUG (service)->priv;
+
+  static const char *configured_caps[] = {
+    IS_CONFIGURED,
+    NULL
+  };
+
+  static const char *authorized_caps[] = {
+    IS_CONFIGURED,
+    CREDENTIALS_VALID,
+    NULL
+  };
+
+  static const char *no_caps[] = { NULL };
+
+  if (priv->authorized)
+    return authorized_caps;
+  else if (priv->configured)
+    return configured_caps;
+  else
+    return no_caps;
+}
+
+static RestXmlNode *
+node_from_call (RestProxyCall *call, GError **error)
+{
+  static RestXmlParser *parser = NULL;
+  RestXmlNode *node;
+
+  if (call == NULL)
+    return NULL;
+
+  if (parser == NULL)
+    parser = rest_xml_parser_new ();
+
+  if (!SOUP_STATUS_IS_SUCCESSFUL (rest_proxy_call_get_status_code (call))) {
+    g_set_error (error, SW_SERVICE_ERROR, SW_SERVICE_ERROR_REMOTE_ERROR,
+                 "HTTP error: %s (%d)",
+                 rest_proxy_call_get_status_message (call),
+                 rest_proxy_call_get_status_code (call));
+    return NULL;
+  }
+
+  node = rest_xml_parser_parse_from_data (parser,
+                                          rest_proxy_call_get_payload (call),
+                                          rest_proxy_call_get_payload_length (call));
+
+  /* Invalid XML, or incorrect root */
+  if (node == NULL || !g_str_equal (node->name, "rsp")) {
+    g_set_error (error, SW_SERVICE_ERROR, SW_SERVICE_ERROR_REMOTE_ERROR,
+                 "malformed remote response: %s",
+                 rest_proxy_call_get_payload (call));
+    if (node)
+      rest_xml_node_unref (node);
+    return NULL;
+  }
+
+  if (g_strcmp0 (rest_xml_node_get_attr (node, "stat"), "ok") != 0) {
+    RestXmlNode *err;
+    err = rest_xml_node_find (node, "err");
+    g_set_error (error, SW_SERVICE_ERROR, SW_SERVICE_ERROR_REMOTE_ERROR,
+                 "remote SmugMug error: %s", err ?
+                 rest_xml_node_get_attr (err, "msg") : "unknown");
+    rest_xml_node_unref (node);
+    return NULL;
+  }
+
+  return node;
+}
+
+static void
+_check_access_token_cb (RestProxyCall *call,
+                        const GError  *error,
+                        GObject       *weak_object,
+                        gpointer       user_data)
+{
+  RestXmlNode *root;
+  GError *err = NULL;
+  SwServiceSmugmug *self = SW_SERVICE_SMUGMUG (weak_object);
+  SwService *service = SW_SERVICE (self);
   SwServiceSmugmugPrivate *priv = self->priv;
 
-  SW_DEBUG (SMUGMUG, "%sauthorized", authorised ? "" : "un");
+  root = node_from_call (call, &err);
+
+  priv->authorized = (root != NULL);
+
+  if (!priv->authorized) {
+    g_assert (err != NULL);
+    SW_DEBUG (SMUGMUG, "Invalid access token: %s", err->message);
+    g_error_free (err);
+  } else {
+    rest_xml_node_unref (root);
+  }
+
+  sw_service_emit_capabilities_changed (service, get_dynamic_caps (service));
+}
+
+static void
+got_tokens_cb (RestProxy *proxy, gboolean got_tokens, gpointer user_data)
+{
+  SwServiceSmugmug *self = (SwServiceSmugmug *) user_data;
+  SwService *service = SW_SERVICE (self);
+  SwServiceSmugmugPrivate *priv = self->priv;
+
+  priv->configured = got_tokens;
+
+  SW_DEBUG (SMUGMUG, "Got tokens: %s", got_tokens ? "yes" : "no");
 
   if (priv->rest_proxy != NULL)
     g_object_unref (priv->rest_proxy);
@@ -117,7 +235,8 @@ got_tokens_cb (RestProxy *proxy, gboolean authorised, gpointer user_data)
   if (priv->upload_proxy != NULL)
     g_object_unref (priv->upload_proxy);
 
-  if (authorised) {
+  if (got_tokens) {
+    RestProxyCall *call;
     const gchar *token = oauth_proxy_get_token ((OAuthProxy *) proxy);
     const gchar *token_secret =
       oauth_proxy_get_token_secret ((OAuthProxy *) proxy);
@@ -135,17 +254,35 @@ got_tokens_cb (RestProxy *proxy, gboolean authorised, gpointer user_data)
                                                      token_secret,
                                                      UPLOAD_URL,
                                                      FALSE);
+
+    call = rest_proxy_new_call (priv->rest_proxy);
+    rest_proxy_call_add_param (call, "method",
+                               "smugmug.auth.checkAccessToken");
+
+    rest_proxy_call_async (call,
+                           _check_access_token_cb,
+                           G_OBJECT (self),
+                           NULL,
+                           NULL);
+
+    g_object_unref (call);
   }
+
+  sw_service_emit_capabilities_changed (service, get_dynamic_caps (service));
 }
 
 static void
 online_notify (gboolean online, gpointer user_data)
 {
   SwServiceSmugmug *self = (SwServiceSmugmug *) user_data;
+  SwService *service = SW_SERVICE (self);
   SwServiceSmugmugPrivate *priv = self->priv;
 
   if (online) {
     sw_keyfob_oauth ((OAuthProxy *) priv->auth_proxy, got_tokens_cb, self);
+  } else {
+    priv->authorized = FALSE;
+    sw_service_emit_capabilities_changed (service, get_dynamic_caps (service));
   }
 }
 
@@ -154,12 +291,19 @@ online_notify (gboolean online, gpointer user_data)
  * the keyring.
  */
 static void
-refresh_credentials (SwServiceSmugmug *smugmug)
+refresh_credentials (SwServiceSmugmug *self)
 {
+  SwService *service = SW_SERVICE (self);
+  SwServiceSmugmugPrivate *priv = self->priv;
+
   SW_DEBUG (SMUGMUG, "Credentials updated");
 
-  online_notify (FALSE, (SwService *) smugmug);
-  online_notify (TRUE, (SwService *) smugmug);
+  priv->configured = FALSE;
+
+  sw_service_emit_user_changed (service);
+
+  online_notify (FALSE, service);
+  online_notify (TRUE, service);
 }
 
 /*
@@ -194,51 +338,6 @@ sw_service_smugmug_dispose (GObject *object)
   G_OBJECT_CLASS (sw_service_smugmug_parent_class)->dispose (object);
 }
 
-static RestXmlNode *
-node_from_call (RestProxyCall *call)
-{
-  static RestXmlParser *parser = NULL;
-  RestXmlNode *node;
-
-  if (call == NULL)
-    return NULL;
-
-  if (parser == NULL)
-    parser = rest_xml_parser_new ();
-
-  if (!SOUP_STATUS_IS_SUCCESSFUL (rest_proxy_call_get_status_code (call))) {
-    g_warning (G_STRLOC ": error from SmugMug: %s (%d)",
-               rest_proxy_call_get_status_message (call),
-               rest_proxy_call_get_status_code (call));
-    return NULL;
-  }
-
-  node = rest_xml_parser_parse_from_data (parser,
-                                          rest_proxy_call_get_payload (call),
-                                          rest_proxy_call_get_payload_length (call));
-  g_object_unref (call);
-
-  /* Invalid XML, or incorrect root */
-  if (node == NULL || !g_str_equal (node->name, "rsp")) {
-    g_warning (G_STRLOC ": cannot make SmugMug call");
-    /* TODO: display the payload if its short */
-    if (node) rest_xml_node_unref (node);
-    return NULL;
-  }
-
-  if (g_strcmp0 (rest_xml_node_get_attr (node, "stat"), "ok") != 0) {
-    RestXmlNode *err;
-    err = rest_xml_node_find (node, "err");
-    if (err)
-      g_warning (G_STRLOC ": cannot make SmugMug call: %s",
-                 rest_xml_node_get_attr (err, "msg"));
-    rest_xml_node_unref (node);
-    return NULL;
-  }
-
-  return node;
-}
-
 static gint
 _upload_file (SwServiceSmugmug *self,
               MediaType upload_type,
@@ -248,16 +347,15 @@ _upload_file (SwServiceSmugmug *self,
               GError **error)
 {
   SwServiceSmugmugPrivate *priv = self->priv;
-  RestProxyCall *call;
+  RestProxyCall *call = NULL;
   RestParam *param;
-  gchar *basename;
-  gchar *content_type;
-  gchar *bytecount;
-  GMappedFile *map;
-  gint opid;
-  GHashTableIter iter;
-  gpointer key, value;
-  GChecksum *checksum;
+  gchar *basename = NULL;
+  gchar *content_type = NULL;
+  gchar *bytecount = NULL;
+  gchar *collection_id = NULL;
+  GMappedFile *map = NULL;
+  gint opid = -1;
+  GChecksum *checksum = NULL;
 
   g_return_val_if_fail (priv->upload_proxy != NULL, -1);
 
@@ -265,7 +363,7 @@ _upload_file (SwServiceSmugmug *self,
   map = g_mapped_file_new (filename, FALSE, error);
   if (*error != NULL) {
     g_warning ("Error opening file %s: %s", filename, (*error)->message);
-    return -1;
+    goto OUT;
   }
 
   /* Get the file information */
@@ -289,30 +387,29 @@ _upload_file (SwServiceSmugmug *self,
   rest_proxy_call_add_param (call, "ResponseType", "REST");
   rest_proxy_call_add_param (call, "ByteCount", bytecount);
 
-  g_hash_table_iter_init (&iter, extra_fields);
-  while (g_hash_table_iter_next (&iter, &key, &value))
-    {
-      const gchar *param_key = key;
-      const gchar *param_value = value;
+  collection_id = g_hash_table_lookup (extra_fields, "collection");
 
-      if (g_strcmp0 (param_key, "collection") == 0)
-        {
-          const gchar *sep = g_strrstr (param_value, "_");
+  if (collection_id == NULL) {
+    g_set_error (error, SW_SERVICE_ERROR, SW_SERVICE_ERROR_NOT_SUPPORTED,
+                 "must provide a collection ID");
+    goto OUT;
+  } else if (!g_str_has_prefix (collection_id, ALBUM_PREFIX) ||
+             g_strstr_len (collection_id, -1, "_") == NULL) {
+    g_set_error (error, SW_SERVICE_ERROR, SW_SERVICE_ERROR_NOT_SUPPORTED,
+                 "collection (%s) must be in the format: %salbumkey_albumid",
+                 collection_id, ALBUM_PREFIX);
+    goto OUT;
+  } else {
+    rest_proxy_call_add_param (call, "AlbumID",
+                               g_strstr_len (collection_id, -1, "_") + 1);
+  }
 
-          if (sep == NULL)
-            g_warning ("'collections' param not valid (%s)", param_value);
-          else
-            rest_proxy_call_add_param (call, "AlbumID", sep + 1);
-        }
-      else if (upload_type != VIDEO && g_strcmp0 (param_key, "title") == 0)
-        {
-          rest_proxy_call_add_param (call, "Caption", param_value);
-        }
-      else if (strlen (param_value) > 0)
-        {
-          rest_proxy_call_add_param (call, param_key, param_value);
-        }
-    }
+
+  sw_service_map_params (upload_params, extra_fields,
+                         (SwServiceSetParamFunc) rest_proxy_call_add_param,
+                         call);
+
+  g_mapped_file_ref (map);
 
   param = rest_param_new_with_owner (basename,
                                      g_mapped_file_get_contents (map),
@@ -336,11 +433,16 @@ _upload_file (SwServiceSmugmug *self,
                          GINT_TO_POINTER (opid),
                          NULL);
 
+ OUT:
   g_free (basename);
   g_free (content_type);
   g_free (bytecount);
-  g_checksum_free (checksum);
-  g_object_unref (call);
+  if (checksum != NULL)
+    g_checksum_free (checksum);
+  if (call != NULL)
+    g_object_unref (call);
+  if (map != NULL)
+    g_mapped_file_unref (map);
 
   return opid;
 }
@@ -417,14 +519,27 @@ _list_albums_cb (RestProxyCall *call,
                  gpointer       user_data)
 {
   DBusGMethodInvocation *context = (DBusGMethodInvocation *) user_data;
-  RestXmlNode *root;
+  RestXmlNode *root = NULL;
   RestXmlNode *album;
-  GPtrArray *rv = g_ptr_array_new_with_free_func (
-      (GDestroyNotify )g_value_array_free);
+  GPtrArray *rv;
+  GError *err = NULL;
 
-  root = node_from_call (call);
+  if (error != NULL)
+    err = g_error_new (SW_SERVICE_ERROR, SW_SERVICE_ERROR_REMOTE_ERROR,
+                       "rest call failed: %s", error->message);
 
-  g_return_if_fail (root != NULL);
+  if (err == NULL)
+    root = node_from_call (call, &err);
+
+  if (err != NULL) {
+    dbus_g_method_return_error (context, err);
+    g_error_free (err);
+    if (root != NULL)
+      rest_xml_node_unref (root);
+    return;
+  }
+
+  rv = g_ptr_array_new_with_free_func ((GDestroyNotify )g_value_array_free);
 
   album = rest_xml_node_find (root, "Album");
 
@@ -469,9 +584,27 @@ _create_album_cb (RestProxyCall *call,
     gpointer       user_data)
 {
   DBusGMethodInvocation *context = (DBusGMethodInvocation *) user_data;
-  RestXmlNode *root = node_from_call (call);
-  RestXmlNode *album = rest_xml_node_find (root, "Album");
+  RestXmlNode *root = NULL;
+  RestXmlNode *album;
   gchar *id;
+  GError *err = NULL;
+
+  if (error != NULL)
+    err = g_error_new (SW_SERVICE_ERROR, SW_SERVICE_ERROR_REMOTE_ERROR,
+                       "rest call failed: %s", error->message);
+
+  if (err == NULL)
+    root = node_from_call (call, &err);
+
+  if (err != NULL) {
+    dbus_g_method_return_error (context, err);
+    g_error_free (err);
+    if (root != NULL)
+      rest_xml_node_unref (root);
+    return;
+  }
+
+  album = rest_xml_node_find (root, "Album");
 
   id = g_strdup_printf ("%s%s_%s", ALBUM_PREFIX,
                         (gchar *) g_hash_table_lookup (album->attrs, "Key"),
@@ -528,8 +661,26 @@ _get_album_details_cb (RestProxyCall *call,
 {
   DBusGMethodInvocation *context = (DBusGMethodInvocation *) user_data;
   GValueArray *collection_details;
-  RestXmlNode *root = node_from_call (call);
-  RestXmlNode *album = rest_xml_node_find (root, "Album");
+  RestXmlNode *root = NULL;
+  RestXmlNode *album;
+  GError *err = NULL;
+
+  if (error != NULL)
+    err = g_error_new (SW_SERVICE_ERROR, SW_SERVICE_ERROR_REMOTE_ERROR,
+                       "rest call failed: %s", error->message);
+
+  if (err == NULL)
+    root = node_from_call (call, &err);
+
+  if (err != NULL) {
+    dbus_g_method_return_error (context, err);
+    g_error_free (err);
+    if (root != NULL)
+      rest_xml_node_unref (root);
+    return;
+  }
+
+  album = rest_xml_node_find (root, "Album");
 
   collection_details = _extract_collection_details_from_xml (album);
 
@@ -553,7 +704,18 @@ _smugmug_collections_get_details (SwCollectionsIface *self,
 
   g_return_if_fail (priv->rest_proxy != NULL);
 
-  id = g_strsplit (collection_id, "_", 2);
+  if (!g_str_has_prefix(collection_id, ALBUM_PREFIX)) {
+    GError *error =
+      g_error_new (SW_SERVICE_ERROR,
+                   SW_SERVICE_ERROR_NOT_SUPPORTED,
+                   "SmugMug collection ID (%s) must start with '%s'",
+                   collection_id, ALBUM_PREFIX);
+    dbus_g_method_return_error (context, error);
+    g_error_free (error);
+    return;
+  }
+
+  id = g_strsplit (collection_id + strlen (ALBUM_PREFIX), "_", 2);
 
   if (g_strv_length (id) != 2) {
     g_warning ("invalid collection id");
@@ -575,6 +737,22 @@ _smugmug_collections_get_details (SwCollectionsIface *self,
   g_object_unref (call);
 }
 
+static void
+_smugmug_collections_get_creatable_types (SwCollectionsIface    *self,
+                                              DBusGMethodInvocation *context)
+{
+  GArray *creatable_types = g_array_sized_new (TRUE, TRUE,
+                                               sizeof (guint), 1);
+  guint v =  PHOTO | VIDEO;
+
+  g_array_append_val (creatable_types, v);
+
+  sw_collections_iface_return_from_get_creatable_types (context,
+                                                        creatable_types);
+
+  g_array_free (creatable_types, TRUE);
+}
+
 static void
 collections_iface_init (gpointer g_iface,
                         gpointer iface_data)
@@ -589,6 +767,9 @@ collections_iface_init (gpointer g_iface,
 
   sw_collections_iface_implement_get_details (klass,
                                               _smugmug_collections_get_details);
+
+  sw_collections_iface_implement_get_creatable_types (klass,
+                                                      _smugmug_collections_get_creatable_types);
 }
 
 /* Photo Upload Interface */
@@ -701,22 +882,53 @@ sw_service_smugmug_class_init (SwServiceSmugmugClass *klass)
 
   service_class->get_name = get_name;
   service_class->get_static_caps = get_static_caps;
+  service_class->get_dynamic_caps = get_dynamic_caps;
   service_class->credentials_updated = credentials_updated;
 }
 
+static gboolean
+sw_service_smugmug_initable (GInitable     *initable,
+                             GCancellable  *cancellable,
+                             GError       **error)
+{
+  SwServiceSmugmug *self = SW_SERVICE_SMUGMUG (initable);
+  SwServiceSmugmugPrivate *priv = self->priv;
+
+  if (priv->inited)
+    return TRUE;
+
+  sw_keystore_get_key_secret ("smugmug", &priv->api_key, &priv->api_secret);
+
+  if (priv->api_key == NULL || priv->api_secret == NULL) {
+    g_set_error_literal (error,
+                         SW_SERVICE_ERROR,
+                         SW_SERVICE_ERROR_NO_KEYS,
+                         "No API or secret key configured");
+    return FALSE;
+  }
+
+  priv->inited = TRUE;
+
+  priv->auth_proxy = oauth_proxy_new (priv->api_key, priv->api_secret,
+                                      OAUTH_URL, FALSE);
+  sw_online_add_notify (online_notify, self);
+  refresh_credentials (self);
+
+  return TRUE;
+}
+
+static void
+initable_iface_init (gpointer g_iface, gpointer iface_data)
+{
+  GInitableIface *klass = (GInitableIface *)g_iface;
+
+  klass->init = sw_service_smugmug_initable;
+}
+
 static void
 sw_service_smugmug_init (SwServiceSmugmug *self)
 {
-  SwServiceSmugmugPrivate *priv;
-
-  priv = self->priv = GET_PRIVATE (self);
-
-  sw_keystore_get_key_secret ("smugmug", &priv->api_key, &priv->api_secret);
-
-  priv->auth_proxy = oauth_proxy_new (priv->api_key, priv->api_secret,
-                                      OAUTH_URL, FALSE);
-
-  sw_online_add_notify (online_notify, self);
-
-  refresh_credentials (self);
+  self->priv = G_TYPE_INSTANCE_GET_PRIVATE (self,
+                                            SW_TYPE_SERVICE_SMUGMUG,
+                                            SwServiceSmugmugPrivate);
 }
